@@ -52,9 +52,11 @@
 //! [`File`]: struct.File.html
 
 use std::fmt::Display;
+use std::num::NonZeroU32;
 use std::ops::Index;
 use std::ops::IndexMut;
 
+use controlled_option::ControlledOption;
 use either::Either;
 use fxhash::FxHashMap;
 use smallvec::SmallVec;
@@ -67,31 +69,31 @@ use crate::arena::SupplementalArena;
 // String content
 
 #[repr(C)]
-struct InternedString {
-    // See InternedStringContent below for how we fill in these fields safely.
+struct InternedStringContent {
+    // See InternedStringArena below for how we fill in these fields safely.
     start: *const u8,
     len: usize,
 }
 
 const INITIAL_STRING_CAPACITY: usize = 512;
 
-/// The content of each `InternedString` is stored in one of the buffers inside of a
-/// `InternedStringContent` instance, following the trick [described by Aleksey Kladov][interner].
+/// The content of each interned string is stored in one of the buffers inside of a
+/// `InternedStringArena` instance, following the trick [described by Aleksey Kladov][interner].
 ///
 /// The buffers stored in this type are preallocated, and are never allowed to grow.  That ensures
 /// that pointers into the buffer are stable, as long as the buffer has not been destroyed.
-/// (`InternedString` instances are also stored in an arena, ensuring that the strings that we hand
-/// out don't outlive the buffers.)
+/// (`InternedStringContent` instances are also stored in an arena, ensuring that the strings that
+/// we hand out don't outlive the buffers.)
 ///
 /// [interner]: https://matklad.github.io/2020/03/22/fast-simple-rust-interner.html
-struct InternedStringContent {
+struct InternedStringArena {
     current_buffer: Vec<u8>,
     full_buffers: Vec<Vec<u8>>,
 }
 
-impl InternedStringContent {
-    fn new() -> InternedStringContent {
-        InternedStringContent {
+impl InternedStringArena {
+    fn new() -> InternedStringArena {
+        InternedStringArena {
             current_buffer: Vec::with_capacity(INITIAL_STRING_CAPACITY),
             full_buffers: Vec::new(),
         }
@@ -99,7 +101,7 @@ impl InternedStringContent {
 
     // Adds a new string.  This does not check whether we've already stored a string with the same
     // content; that is handled down below in `StackGraph::add_symbol` and `add_file`.
-    fn add(&mut self, value: &str) -> InternedString {
+    fn add(&mut self, value: &str) -> InternedStringContent {
         // Is there enough room in current_buffer to hold this string?
         let value = value.as_bytes();
         let len = value.len();
@@ -121,15 +123,15 @@ impl InternedStringContent {
         let start_index = self.current_buffer.len();
         self.current_buffer.extend_from_slice(value);
         let start = unsafe { self.current_buffer.as_ptr().add(start_index) };
-        InternedString { start, len }
+        InternedStringContent { start, len }
     }
 }
 
-impl InternedString {
+impl InternedStringContent {
     /// Returns the content of this string as a `str`.  This is safe as long as the lifetime of the
-    /// InternedString is outlived by the lifetime of the InternedStringContent that holds its
+    /// InternedStringContent is outlived by the lifetime of the InternedStringArena that holds its
     /// data.  That is guaranteed because we store the InternedStrings in an Arena alongside the
-    /// InternedStringContent, and only hand out references to them.
+    /// InternedStringArena, and only hand out references to them.
     fn as_str(&self) -> &str {
         unsafe {
             let bytes = std::slice::from_raw_parts(self.start, self.len);
@@ -140,7 +142,7 @@ impl InternedString {
     // Returns a supposedly 'static reference to the string's data.  The string data isn't really
     // static, but we are careful only to use this as a key in the HashMap that StackGraph uses to
     // track whether we've stored a particular symbol already.  That HashMap lives alongside the
-    // InternedStringContent that holds the data, so we can get away with a technically incorrect
+    // InternedStringArena that holds the data, so we can get away with a technically incorrect
     // 'static lifetime here.  As an extra precaution, this method is is marked as unsafe so that
     // we don't inadvertently call it from anywhere else in the crate.
     unsafe fn as_hash_key(&self) -> &'static str {
@@ -164,7 +166,7 @@ impl InternedString {
 /// to symbols using simple equality, without having to dereference into the `StackGraph` arena.
 #[repr(C)]
 pub struct Symbol {
-    content: InternedString,
+    content: InternedStringContent,
 }
 
 impl Symbol {
@@ -232,6 +234,80 @@ impl Handle<Symbol> {
     }
 }
 
+//-------------------------------------------------------------------------------------------------
+// Interned strings
+
+/// Arbitrary string content associated with some part of a stack graph.
+#[repr(C)]
+pub struct InternedString {
+    content: InternedStringContent,
+}
+
+impl InternedString {
+    fn as_str(&self) -> &str {
+        self.content.as_str()
+    }
+}
+
+impl PartialEq<&str> for InternedString {
+    fn eq(&self, other: &&str) -> bool {
+        self.as_str() == *other
+    }
+}
+
+impl StackGraph {
+    /// Adds an interned string to the stack graph, ensuring that there's only ever one copy of a
+    /// particular string stored in the graph.
+    pub fn add_string<S: AsRef<str> + ?Sized>(&mut self, string: &S) -> Handle<InternedString> {
+        let string = string.as_ref();
+        if let Some(handle) = self.string_handles.get(string) {
+            return *handle;
+        }
+
+        let interned = self.interned_strings.add(string);
+        let hash_key = unsafe { interned.as_hash_key() };
+        let handle = self.strings.add(InternedString { content: interned });
+        self.string_handles.insert(hash_key, handle);
+        handle
+    }
+
+    /// Returns an iterator over all of the handles of all of the interned strings in this stack
+    /// graph. (Note that because we're only returning _handles_, this iterator does not retain a
+    /// reference to the `StackGraph`.)
+    pub fn iter_strings(&self) -> impl Iterator<Item = Handle<InternedString>> {
+        self.strings.iter_handles()
+    }
+}
+
+impl Index<Handle<InternedString>> for StackGraph {
+    type Output = str;
+    #[inline(always)]
+    fn index(&self, handle: Handle<InternedString>) -> &str {
+        self.strings.get(handle).as_str()
+    }
+}
+
+#[doc(hidden)]
+pub struct DisplayInternedString<'a> {
+    wrapped: Handle<InternedString>,
+    graph: &'a StackGraph,
+}
+
+impl<'a> Display for DisplayInternedString<'a> {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        write!(f, "{}", &self.graph[self.wrapped])
+    }
+}
+
+impl Handle<InternedString> {
+    pub fn display(self, graph: &StackGraph) -> impl Display + '_ {
+        DisplayInternedString {
+            wrapped: self,
+            graph,
+        }
+    }
+}
+
 //-------------------------------------------------------------------------------------------------
 // Files
 
@@ -244,7 +320,7 @@ impl Handle<Symbol> {
 /// each other.
 pub struct File {
     /// The name of this source file.
-    name: InternedString,
+    name: InternedStringContent,
 }
 
 impl File {
@@ -350,7 +426,7 @@ impl Handle<File> {
 #[repr(C)]
 #[derive(Clone, Copy, Default, Eq, Hash, Ord, PartialEq, PartialOrd)]
 pub struct NodeID {
-    file: Option<Handle<File>>,
+    file: ControlledOption<Handle<File>>,
     local_id: u32,
 }
 
@@ -362,7 +438,7 @@ impl NodeID {
     #[inline(always)]
     pub fn root() -> NodeID {
         NodeID {
-            file: None,
+            file: ControlledOption::none(),
             local_id: ROOT_NODE_ID,
         }
     }
@@ -371,7 +447,7 @@ impl NodeID {
     #[inline(always)]
     pub fn jump_to() -> NodeID {
         NodeID {
-            file: None,
+            file: ControlledOption::none(),
             local_id: JUMP_TO_NODE_ID,
         }
     }
@@ -380,7 +456,7 @@ impl NodeID {
     #[inline(always)]
     pub fn new_in_file(file: Handle<File>, local_id: u32) -> NodeID {
         NodeID {
-            file: Some(file),
+            file: ControlledOption::some(file),
             local_id,
         }
     }
@@ -401,7 +477,7 @@ impl NodeID {
     /// _jump to scope_ nodes, which belong to all files.
     #[inline(always)]
     pub fn file(self) -> Option<Handle<File>> {
-        self.file
+        self.file.into()
     }
 
     /// Returns the local ID of this node within its file.  Panics if this node ID refers to the
@@ -416,7 +492,7 @@ impl NodeID {
     /// singleton _root_ and _jump to scope_ nodes, which belong to all files.
     #[inline(always)]
     pub fn is_in_file(self, file: Handle<File>) -> bool {
-        match self.file {
+        match self.file.into_option() {
             Some(this_file) => this_file == file,
             _ => true,
         }
@@ -431,7 +507,7 @@ pub struct DisplayNodeID<'a> {
 
 impl<'a> Display for DisplayNodeID<'a> {
     fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
-        match self.wrapped.file {
+        match self.wrapped.file.into_option() {
             Some(file) => write!(f, "{}({})", file.display(self.graph), self.wrapped.local_id),
             None => {
                 if self.wrapped.is_root() {
@@ -525,12 +601,12 @@ impl Node {
             Node::DropScopes(node) => node.id,
             Node::ExportedScope(node) => node.id,
             Node::InternalScope(node) => node.id,
-            Node::JumpTo(_) => NodeID::jump_to(),
+            Node::JumpTo(node) => node.id,
             Node::PushScopedSymbol(node) => node.id,
             Node::PushSymbol(node) => node.id,
             Node::PopScopedSymbol(node) => node.id,
             Node::PopSymbol(node) => node.id,
-            Node::Root(_) => NodeID::root(),
+            Node::Root(node) => node.id,
         }
     }
 
@@ -559,14 +635,14 @@ impl Node {
 impl StackGraph {
     /// Returns a handle to the stack graph's singleton _jump to scope_ node.
     #[inline(always)]
-    pub fn jump_to_node(&self) -> Handle<Node> {
-        self.jump_to_node
+    pub fn jump_to_node() -> Handle<Node> {
+        Handle::new(unsafe { NonZeroU32::new_unchecked(2) })
     }
 
     /// Returns a handle to the stack graph's singleton _root node_.
     #[inline(always)]
-    pub fn root_node(&self) -> Handle<Node> {
-        self.root_node
+    pub fn root_node() -> Handle<Node> {
+        Handle::new(unsafe { NonZeroU32::new_unchecked(1) })
     }
 
     /// Returns an unused [`NodeID`][] for the given file.
@@ -587,9 +663,9 @@ impl StackGraph {
         if id.file().is_some() {
             self.node_id_handles.try_handle_for_id(id)
         } else if id.is_root() {
-            Some(self.root_node())
+            Some(StackGraph::root_node())
         } else if id.is_jump_to() {
-            Some(self.jump_to_node())
+            Some(StackGraph::jump_to_node())
         } else {
             None
         }
@@ -603,6 +679,15 @@ impl StackGraph {
         self.node_id_handles.set_handle_for_id(id, handle);
         Some(handle)
     }
+
+    pub(crate) fn get_or_create_node(&mut self, id: NodeID, node: Node) -> Handle<Node> {
+        if let Some(handle) = self.node_id_handles.handle_for_id(id) {
+            return handle;
+        }
+        let handle = self.nodes.add(node);
+        self.node_id_handles.set_handle_for_id(id, handle);
+        handle
+    }
 }
 
 #[doc(hidden)]
@@ -656,7 +741,7 @@ impl IndexMut<Handle<Node>> for StackGraph {
 pub struct DropScopesNode {
     /// The unique identifier for this node.
     pub id: NodeID,
-    _symbol: Option<Handle<Symbol>>,
+    _symbol: ControlledOption<Handle<Symbol>>,
     _scope: NodeID,
     _is_clickable: bool,
 }
@@ -672,7 +757,7 @@ impl StackGraph {
     pub fn add_drop_scopes_node(&mut self, id: NodeID) -> Option<Handle<Node>> {
         let node = DropScopesNode {
             id,
-            _symbol: None,
+            _symbol: ControlledOption::none(),
             _scope: NodeID::default(),
             _is_clickable: false,
         };
@@ -711,7 +796,7 @@ impl<'a> Display for DisplayDropScopesNode<'a> {
 pub struct ExportedScopeNode {
     /// The unique identifier for this node.
     pub id: NodeID,
-    _symbol: Option<Handle<Symbol>>,
+    _symbol: ControlledOption<Handle<Symbol>>,
     _scope: NodeID,
     _is_clickable: bool,
 }
@@ -727,7 +812,7 @@ impl StackGraph {
     pub fn add_exported_scope_node(&mut self, id: NodeID) -> Option<Handle<Node>> {
         let node = ExportedScopeNode {
             id,
-            _symbol: None,
+            _symbol: ControlledOption::none(),
             _scope: NodeID::default(),
             _is_clickable: false,
         };
@@ -770,7 +855,7 @@ impl<'a> Display for DisplayExportedScopeNode<'a> {
 pub struct InternalScopeNode {
     /// The unique identifier for this node.
     pub id: NodeID,
-    _symbol: Option<Handle<Symbol>>,
+    _symbol: ControlledOption<Handle<Symbol>>,
     _scope: NodeID,
     _is_clickable: bool,
 }
@@ -786,7 +871,7 @@ impl StackGraph {
     pub fn add_internal_scope_node(&mut self, id: NodeID) -> Option<Handle<Node>> {
         let node = InternalScopeNode {
             id,
-            _symbol: None,
+            _symbol: ControlledOption::none(),
             _scope: NodeID::default(),
             _is_clickable: false,
         };
@@ -827,8 +912,8 @@ impl<'a> Display for DisplayInternalScopeNode<'a> {
 /// the graph.
 #[repr(C)]
 pub struct JumpToNode {
-    _id: NodeID,
-    _symbol: Option<Handle<Symbol>>,
+    id: NodeID,
+    _symbol: ControlledOption<Handle<Symbol>>,
     _scope: NodeID,
     _is_clickable: bool,
 }
@@ -842,11 +927,8 @@ impl From<JumpToNode> for Node {
 impl JumpToNode {
     fn new() -> JumpToNode {
         JumpToNode {
-            _id: NodeID {
-                file: None,
-                local_id: 0,
-            },
-            _symbol: None,
+            id: NodeID::jump_to(),
+            _symbol: ControlledOption::none(),
             _scope: NodeID::default(),
             _is_clickable: false,
         }
@@ -1155,8 +1237,8 @@ impl<'a> Display for DisplayPushSymbolNode<'a> {
 /// The singleton root node, which allows a name binding path to cross between files.
 #[repr(C)]
 pub struct RootNode {
-    _id: NodeID,
-    _symbol: Option<Handle<Symbol>>,
+    id: NodeID,
+    _symbol: ControlledOption<Handle<Symbol>>,
     _scope: NodeID,
     _is_clickable: bool,
 }
@@ -1170,11 +1252,8 @@ impl From<RootNode> for Node {
 impl RootNode {
     fn new() -> RootNode {
         RootNode {
-            _id: NodeID {
-                file: None,
-                local_id: 0,
-            },
-            _symbol: None,
+            id: NodeID::root(),
+            _symbol: ControlledOption::none(),
             _scope: NodeID::default(),
             _is_clickable: false,
         }
@@ -1199,7 +1278,7 @@ impl NodeIDHandles {
     }
 
     fn try_handle_for_id(&self, node_id: NodeID) -> Option<Handle<Node>> {
-        let file_entry = self.files.get(node_id.file.unwrap())?;
+        let file_entry = self.files.get(node_id.file().unwrap())?;
         let node_index = node_id.local_id as usize;
         if node_index >= file_entry.len() {
             return None;
@@ -1208,7 +1287,7 @@ impl NodeIDHandles {
     }
 
     fn handle_for_id(&mut self, node_id: NodeID) -> Option<Handle<Node>> {
-        let file_entry = &mut self.files[node_id.file.unwrap()];
+        let file_entry = &mut self.files[node_id.file().unwrap()];
         let node_index = node_id.local_id as usize;
         if node_index >= file_entry.len() {
             file_entry.resize(node_index + 1, None);
@@ -1217,7 +1296,7 @@ impl NodeIDHandles {
     }
 
     fn set_handle_for_id(&mut self, node_id: NodeID, handle: Handle<Node>) {
-        let file_entry = &mut self.files[node_id.file.unwrap()];
+        let file_entry = &mut self.files[node_id.file().unwrap()];
         let node_index = node_id.local_id as usize;
         file_entry[node_index] = Some(handle);
     }
@@ -1291,20 +1370,49 @@ impl StackGraph {
     }
 }
 
+//-------------------------------------------------------------------------------------------------
+// Source code
+
+/// Contains information about a range of code in a source code file.
+#[repr(C)]
+#[derive(Default)]
+pub struct SourceInfo {
+    /// The location in its containing file of the source code that this node represents.
+    pub span: lsp_positions::Span,
+    /// The kind of syntax entity this node represents (e.g. `function`, `class`, `method`, etc.).
+    pub syntax_type: Option<Handle<InternedString>>,
+    /// The full content of the line containing this node in its source file.
+    pub containing_line: ControlledOption<Handle<InternedString>>,
+}
+
+impl StackGraph {
+    /// Returns information about the source code that a stack graph node represents.
+    pub fn source_info(&self, node: Handle<Node>) -> Option<&SourceInfo> {
+        self.source_info.get(node)
+    }
+
+    /// Returns a mutable reference to the information about the source code that a stack graph
+    /// node represents.
+    pub fn source_info_mut(&mut self, node: Handle<Node>) -> &mut SourceInfo {
+        &mut self.source_info[node]
+    }
+}
+
 //-------------------------------------------------------------------------------------------------
 // Stack graphs
 
 /// Contains all of the nodes and edges that make up a stack graph.
 pub struct StackGraph {
-    interned_strings: InternedStringContent,
+    interned_strings: InternedStringArena,
     pub(crate) symbols: Arena<Symbol>,
     symbol_handles: FxHashMap<&'static str, Handle<Symbol>>,
+    pub(crate) strings: Arena<InternedString>,
+    string_handles: FxHashMap<&'static str, Handle<InternedString>>,
     pub(crate) files: Arena<File>,
     file_handles: FxHashMap<&'static str, Handle<File>>,
     pub(crate) nodes: Arena<Node>,
+    pub(crate) source_info: SupplementalArena<Node, SourceInfo>,
     node_id_handles: NodeIDHandles,
-    jump_to_node: Handle<Node>,
-    root_node: Handle<Node>,
     outgoing_edges: SupplementalArena<Node, SmallVec<[OutgoingEdge; 8]>>,
 }
 
@@ -1318,19 +1426,20 @@ impl StackGraph {
 impl Default for StackGraph {
     fn default() -> StackGraph {
         let mut nodes = Arena::new();
-        let root_node = nodes.add(RootNode::new().into());
-        let jump_to_node = nodes.add(JumpToNode::new().into());
+        nodes.add(RootNode::new().into());
+        nodes.add(JumpToNode::new().into());
 
         StackGraph {
-            interned_strings: InternedStringContent::new(),
+            interned_strings: InternedStringArena::new(),
             symbols: Arena::new(),
             symbol_handles: FxHashMap::default(),
+            strings: Arena::new(),
+            string_handles: FxHashMap::default(),
             files: Arena::new(),
             file_handles: FxHashMap::default(),
             nodes,
+            source_info: SupplementalArena::new(),
             node_id_handles: NodeIDHandles::new(),
-            jump_to_node,
-            root_node,
             outgoing_edges: SupplementalArena::new(),
         }
     }
