@@ -5,6 +5,7 @@
 // Please see the LICENSE-APACHE or LICENSE-MIT files in this distribution for license details.
 // ------------------------------------------------------------------------------------------------
 
+use controlled_option::ControlledOption;
 use libc::c_char;
 use stack_graphs::arena::Handle;
 use stack_graphs::c::sg_file_handle;
@@ -13,6 +14,7 @@ use stack_graphs::c::sg_stack_graph_add_files;
 use stack_graphs::c::sg_stack_graph_files;
 use stack_graphs::c::sg_stack_graph_free;
 use stack_graphs::c::sg_stack_graph_new;
+use stack_graphs::c::SG_NULL_HANDLE;
 use stack_graphs::graph::File;
 
 fn lengths(data: &[&'static str]) -> Vec<usize> {
@@ -66,3 +68,13 @@ fn can_create_files() {
 
     sg_stack_graph_free(graph);
 }
+
+#[test]
+#[allow(unused_assignments)]
+fn verify_null_file_representation() {
+    let bytes = [0x55u8; std::mem::size_of::<Handle<File>>()];
+    let mut rust: ControlledOption<Handle<File>> = unsafe { std::mem::transmute(bytes) };
+    rust = ControlledOption::none();
+    let c: sg_file_handle = unsafe { std::mem::transmute(rust) };
+    assert_eq!(c, SG_NULL_HANDLE);
+}
