@@ -5,6 +5,7 @@
 // Please see the LICENSE-APACHE or LICENSE-MIT files in this distribution for license details.
 // ------------------------------------------------------------------------------------------------
 
+use controlled_option::ControlledOption;
 use either::Either;
 use libc::c_char;
 use stack_graphs::c::sg_deque_direction;
@@ -29,18 +30,22 @@ use stack_graphs::c::sg_scope_stack_cells;
 use stack_graphs::c::sg_scoped_symbol;
 use stack_graphs::c::sg_stack_graph;
 use stack_graphs::c::sg_stack_graph_add_files;
-use stack_graphs::c::sg_stack_graph_add_nodes;
 use stack_graphs::c::sg_stack_graph_add_symbols;
 use stack_graphs::c::sg_stack_graph_free;
+use stack_graphs::c::sg_stack_graph_get_or_create_nodes;
 use stack_graphs::c::sg_stack_graph_new;
 use stack_graphs::c::sg_symbol_handle;
 use stack_graphs::c::sg_symbol_stack;
 use stack_graphs::c::sg_symbol_stack_cells;
 use stack_graphs::c::SG_LIST_EMPTY_HANDLE;
+use stack_graphs::c::SG_NULL_HANDLE;
+use stack_graphs::paths::PathEdgeList;
+use stack_graphs::paths::ScopeStack;
+use stack_graphs::paths::SymbolStack;
 
 fn add_file(graph: *mut sg_stack_graph, filename: &str) -> sg_file_handle {
     let lengths = [filename.len()];
-    let mut handles: [sg_file_handle; 1] = [0; 1];
+    let mut handles: [sg_file_handle; 1] = [SG_NULL_HANDLE; 1];
     sg_stack_graph_add_files(
         graph,
         1,
@@ -48,13 +53,13 @@ fn add_file(graph: *mut sg_stack_graph, filename: &str) -> sg_file_handle {
         lengths.as_ptr(),
         handles.as_mut_ptr(),
     );
-    assert!(handles[0] != 0);
+    assert!(handles[0] != SG_NULL_HANDLE);
     handles[0]
 }
 
 fn add_symbol(graph: *mut sg_stack_graph, value: &str) -> sg_symbol_handle {
     let lengths = [value.len()];
-    let mut handles: [sg_symbol_handle; 1] = [0; 1];
+    let mut handles: [sg_symbol_handle; 1] = [SG_NULL_HANDLE; 1];
     sg_stack_graph_add_symbols(
         graph,
         1,
@@ -62,7 +67,7 @@ fn add_symbol(graph: *mut sg_stack_graph, value: &str) -> sg_symbol_handle {
         lengths.as_ptr(),
         handles.as_mut_ptr(),
     );
-    assert!(handles[0] != 0);
+    assert!(handles[0] != SG_NULL_HANDLE);
     handles[0]
 }
 
@@ -74,13 +79,13 @@ fn add_exported_scope(
     let node = sg_node {
         kind: sg_node_kind::SG_NODE_KIND_EXPORTED_SCOPE,
         id: sg_node_id { file, local_id },
-        symbol: 0,
+        symbol: SG_NULL_HANDLE,
         is_clickable: false,
         scope: sg_node_id::default(),
     };
     let nodes = [node];
-    let mut handles: [sg_node_handle; 1] = [0; 1];
-    sg_stack_graph_add_nodes(graph, nodes.len(), nodes.as_ptr(), handles.as_mut_ptr());
+    let mut handles: [sg_node_handle; 1] = [SG_NULL_HANDLE; 1];
+    sg_stack_graph_get_or_create_nodes(graph, nodes.len(), nodes.as_ptr(), handles.as_mut_ptr());
     handles[0]
 }
 
@@ -88,7 +93,10 @@ fn add_exported_scope(
 // Symbol stacks
 
 fn empty_scope_stack() -> sg_scope_stack {
-    sg_scope_stack { cells: 0 }
+    sg_scope_stack {
+        cells: SG_NULL_HANDLE,
+        length: 0,
+    }
 }
 
 fn scoped_symbol(symbol: sg_symbol_handle, scopes: sg_scope_stack) -> sg_scoped_symbol {
@@ -171,6 +179,16 @@ fn can_create_symbol_stacks() {
     sg_stack_graph_free(graph);
 }
 
+#[test]
+#[allow(unused_assignments)]
+fn verify_null_symbol_stack_representation() {
+    let bytes = [0x55u8; std::mem::size_of::<SymbolStack>()];
+    let mut rust: ControlledOption<SymbolStack> = unsafe { std::mem::transmute(bytes) };
+    rust = ControlledOption::none();
+    let c: sg_symbol_stack = unsafe { std::mem::transmute(rust) };
+    assert_eq!(c.cells, SG_NULL_HANDLE);
+}
+
 //-------------------------------------------------------------------------------------------------
 // Scope stacks
 
@@ -232,6 +250,16 @@ fn can_create_scope_stacks() {
     sg_stack_graph_free(graph);
 }
 
+#[test]
+#[allow(unused_assignments)]
+fn verify_null_scope_stack_representation() {
+    let bytes = [0x55u8; std::mem::size_of::<ScopeStack>()];
+    let mut rust: ControlledOption<ScopeStack> = unsafe { std::mem::transmute(bytes) };
+    rust = ControlledOption::none();
+    let c: sg_scope_stack = unsafe { std::mem::transmute(rust) };
+    assert_eq!(c.cells, SG_NULL_HANDLE);
+}
+
 //-------------------------------------------------------------------------------------------------
 // Path edge lists
 
@@ -278,7 +306,7 @@ fn path_edge_list_available_in_both_directions(
         return true;
     }
     let cell = &cells[head as usize];
-    cell.reversed != 0
+    cell.reversed != SG_NULL_HANDLE
 }
 
 #[test]
@@ -329,3 +357,13 @@ fn can_create_path_edge_lists() {
     sg_path_arena_free(paths);
     sg_stack_graph_free(graph);
 }
+
+#[test]
+#[allow(unused_assignments)]
+fn verify_null_path_edge_list_representation() {
+    let bytes = [0x55u8; std::mem::size_of::<PathEdgeList>()];
+    let mut rust: ControlledOption<PathEdgeList> = unsafe { std::mem::transmute(bytes) };
+    rust = ControlledOption::none();
+    let c: sg_path_edge_list = unsafe { std::mem::transmute(rust) };
+    assert_eq!(c.cells, SG_NULL_HANDLE);
+}
